# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import evoflow.backend.tensorflow as B

# from evoflow.backend.tf_ops.assign import assign as assign2

import tensorflow as tf
from perfcounters import PerfCounters


@tf.function()
def randint_tf(low, high, shape):
    return B.randint(low, high, shape)


@tf.function(experimental_compile=True)
def randint_xla(low, high, shape):
    return B.randint(low, high, shape)


def test_randint_fn():
    NUM_TESTS = 3
    low = 10
    high = 1000
    shape = (100, 100, 100)

    randint_tf(low, high, shape)
    randint_xla(low, high, shape)

    v = randint_tf(low, high, shape)
    v2 = randint_tf(low, high, shape)
    assert not B.tensor_equal(v, v2)

    v = randint_xla(low, high, shape)
    v2 = randint_xla(low, high, shape)
    assert not B.tensor_equal(v, v2)

    cnts = PerfCounters()
    cnts.start('basic')
    for _ in range(NUM_TESTS):
        B.randint(low, high, shape)
    cnts.stop('basic')

    cnts.start('tf_fn')
    for _ in range(NUM_TESTS):
        randint_tf(low, high, shape)
    cnts.stop('tf_fn')

    cnts.start('tf_xla')
    for _ in range(NUM_TESTS):
        randint_xla(low, high, shape)
    cnts.stop('tf_xla')

    cnts.report()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from termcolor import cprint
from perfcounters import PerfCounters
from tqdm.auto import tqdm
import inspect


def slices2array(slices):
    "convert slices into an array. Use for assign"
    return [[s.start, s.stop] for s in slices]


def box(a):
    "box single tensor into a list"
    if isinstance(a, list):
        return a
    else:
        return [a]


def unbox(a):
    "Unbox list with a single element"
    if isinstance(a, list) and len(a) == 1:
        return a[0]
    else:
        return a


def op_optimization_benchmark(population, op, num_runs, num_warmup=1):
    """Run OP with the different level of optimization supported and
    report results as PerfCounters counterts.

    Args:
        population (tensor): population to run the bench on
        op (OP): OP to be benchmarked
        num_runs (int): number of execution to perform.

    Returns:
        PerfCounters: performance counters for various optimization levels.
    """
    cprint('[warn-up]', 'yellow')
    os = optimization_support(op)
    max_level = os['optimization_level']
    cprint('Max optimization level %s' % (max_level), 'magenta')
    cprint('Optimization status:%s' % os['optimizations'], 'blue')

    cprint('[%s micro benchmark]' % str(op.__class__.__name__), 'yellow')

    total_tests = (num_warmup + num_runs) * (max_level + 1)
    pb = tqdm(total=total_tests, desc='running tests', unit='ops')
    cnts = PerfCounters()

    for level in range(max_level + 1):
        cname = 'Optimization level: %d' % level
        op.set_optimization_level(level)

        # warmup
        for _ in range(num_warmup):
            op(population)
            pb.update(1)

        # real test
        cnts.start(cname)
        for _ in range(num_runs):
            op(population)
            pb.update(1)
        cnts.stop(cname)

    pb.close()
    return cnts


def optimization_support(op):
    """Return the level of optimization supported by a given op and
       which optimizations are supported

    Args:
        op (OP): the op to analyze

    Returns:
        dict: optimization info.
    """

    optims = {}
    for m in inspect.getmembers(op):
        if 'O_' in m[0]:
            optims[m[0]] = m[1]

    autograph = optims.get('O_AUTOGRAPH', False)
    xla = optims.get('O_XLA', False)

    if autograph and xla:
        max_level = 2
    elif autograph:
        max_level = 1
    else:
        max_level = 0

    return {
        'optimization_level': max_level,
        'optimizations': {
            'autograph': autograph,
            'xla': xla
        }
    }



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from termcolor import cprint
import tensorflow as tf

# ! MUST BE IN ITS OWN FILES TO BE TRACEABLE - DO NOT MOVE!
# https://github.com/tensorflow/tensorflow/issues/30409
# https://github.com/tensorflow/tensorflow/issues/34683


# only working version so far
def assign(dst_tensor, updates, indexes):
    """ assign values in tensor at the position specified by the slices.

    Args:
        dst_tensor (ndarray): Target tensor
        values (ndarray): Tensor containing the values to assign.
        indexes (list(list)): python array of the form:
        [[start][stop], ..., [start][stop]] where len == rank of dst_tensor.
        MUST be a python array not a tensor.

    Returns:
        ndarray: tensor with the assigned values.

    """
    # yes this code is verbose and can be simplified.. however its easier to
    # keep each case separated for debug and optim test idea purpose so leaving
    # it as in autograph simplifies it

    num_dims = len(indexes)
    updates = tf.reshape(updates, [-1])

    #  array size -> doesn't seems to provide any benefits
    arr_size = 1
    for d in indexes:
        arr_size *= (d[1] - d[0])

    v = 0
    indices = tf.TensorArray(dtype=tf.int32,
                             size=arr_size,
                             dynamic_size=False,
                             element_shape=(num_dims, ))

    if num_dims == 2:
        for d1 in range(indexes[0][0], indexes[0][1]):
            for d2 in range(indexes[1][0], indexes[1][1]):
                indices = indices.write(v, [d1, d2])
                v += 1

    elif num_dims == 3:
        for d1 in tf.range(indexes[0][0], indexes[0][1]):
            for d2 in tf.range(indexes[1][0], indexes[1][1]):
                for d3 in tf.range(indexes[2][0], indexes[2][1]):
                    indices = indices.write(v, [d1, d2, d3])
                    v += 1

    elif num_dims == 4:
        for d1 in tf.range(indexes[0][0], indexes[0][1]):
            for d2 in tf.range(indexes[1][0], indexes[1][1]):
                for d3 in tf.range(indexes[2][0], indexes[2][1]):
                    for d4 in tf.range(indexes[3][0], indexes[3][1]):
                        indices = indices.write(v, [d1, d2, d3, d4])
                        v += 1
    else:
        raise ValueError('Not implemented for rank >2')

    indices = indices.stack()  # to tensor
    return tf.tensor_scatter_nd_update(dst_tensor,
                                       indices=indices,
                                       updates=updates)


# ! not working tensor placement error again
def assign_batched_failed(dst_tensor, updates, indexes):
    """ assign values in tensor at the position specified by the slices.

    Args:
        dst_tensor (ndarray): Target tensor
        values (ndarray): Tensor containing the values to assign.
        indexes (list(list)): python array of the form:
        [[start][stop], ..., [start][stop]] where len == rank of dst_tensor.
        MUST be a python array not a tensor.

    Returns:
        ndarray: tensor with the assigned values.

    """
    # yes this code is verbose and can be simplified.. however its easier to
    # keep each case separated for debug and testing optim ideas purpose
    # it as in autograph simplifies it

    num_dims = len(indexes)
    updates = tf.reshape(updates, [-1])

    #  array size
    arr_size = 1
    for d in indexes[:-1]:
        arr_size *= (d[1] - d[0])

    v = 0
    indices = tf.TensorArray(dtype=tf.int32, size=arr_size, dynamic_size=False)
    # ,element_shape=(num_dims * last_dim_size, ))

    if num_dims == 3:
        for d1 in tf.range(indexes[0][0], indexes[0][1]):
            for d2 in tf.range(indexes[1][0], indexes[1][1]):

                # try to minimize the update by batching it
                update = []
                for d3 in range(indexes[2][0], indexes[2][1]):
                    update.append([d1, d2, d3])

                indices = indices.write(v, update)
                v += 1
    else:
        raise ValueError('Not implemented for rank 1 and >4')

    indices = indices.stack()  # to tensor
    cprint(indices, 'yellow')
    cprint(indices.shape, 'blue')
    return tf.tensor_scatter_nd_update(dst_tensor,
                                       indices=indices,
                                       updates=updates)


# slower
def assign2(dst_tensor, updates, indexes):
    """ assign values in tensor at the position specified by the slices.

    Args:
        dst_tensor (ndarray): Target tensor
        values (ndarray): Tensor containing the values to assign.
        indexes (list(list)): python array of the form:
        [[start][stop], ..., [start][stop]] where len == rank of dst_tensor.
        MUST be a python array not a tensor.

    Returns:
        ndarray: tensor with the assigned values.

    """
    # yes this code is verbose and can be simplified.. however its easier to
    # keep each case separated for debug and optim test idea purpose so leaving
    # it as in autograph simplifies it

    num_dims = len(indexes)
    updates = tf.reshape(updates, [-1])

    # passing arr_size is also not working
    # Not working
    #  array size
    # arr_size = 1
    # for d in indexes:
    #     arr_size *= (d[1] - d[0])
    # print(arr_size)
    if num_dims == 2:
        indices = tf.TensorArray(dtype=tf.int32,
                                 size=0,
                                 dynamic_size=True,
                                 element_shape=(1, ))
        for d1idx in range(indexes[0][0], indexes[0][1]):
            for d2idx in range(indexes[1][0], indexes[1][1]):
                indices = indices.write(indices.size(), d1idx)
                indices = indices.write(indices.size(), d2idx)

    elif num_dims == 3:
        indices = tf.TensorArray(dtype=tf.int32, size=0, dynamic_size=True)
        for d1idx in tf.range(indexes[0][0], indexes[0][1]):
            for d2idx in tf.range(indexes[1][0], indexes[1][1]):
                for d3idx in tf.range(indexes[2][0], indexes[2][1]):
                    indices = indices.write(indices.size(), d1idx)
                    indices = indices.write(indices.size(), d2idx)
                    indices = indices.write(indices.size(), d3idx)
    else:
        raise ValueError('Not implemented for rank >2')

    arr_len = indices.size()
    indices = indices.stack()
    dim_size = tf.cast(arr_len / num_dims, 'int32')
    indices = tf.reshape(indices, (dim_size, num_dims))
    # cprint(arr_len, 'blue')
    # cprint(dim_size, 'yellow')
    # cprint(indices.shape, 'green')

    return tf.tensor_scatter_nd_update(dst_tensor,
                                       indices=indices,
                                       updates=updates)


# not working with randint but wish it would
def assign_broken(dst_tensor, updates, indexes):
    """ assign values in tensor at the position specified by the slices.

    Args:
        dst_tensor (ndarray): Target tensor
        values (ndarray): Tensor containing the values to assign.
        indexes (list(list)): python array of the form:
        [[start][stop], ..., [start][stop]] where len == rank of dst_tensor.
        MUST be a python array not a tensor.

    Returns:
        ndarray: tensor with the assigned values.

    """
    # !gotcha:
    # ! - slices must be a python array
    # ! - do not use tf.range
    # ! - need its own file (why???)

    # yes this code is verbose and can be simplified.. however its easier to
    # keep each case separated for debug and optim test idea purpose so leaving
    # it as in autograph simplifies it

    num_dims = len(indexes)
    updates = tf.reshape(updates, [-1])

    indices = []
    if num_dims == 1:
        for idx in range(indexes[0][0], indexes[0][1]):
            indices.append(idx)

    elif num_dims == 2:
        for d1idx in range(indexes[0][0], indexes[0][1]):
            for d2idx in range(indexes[1][0], indexes[1][1]):
                indices.append([d1idx, d2idx])
    elif num_dims == 3:
        for d1idx in range(indexes[0][0], indexes[0][1]):
            for d2idx in range(indexes[1][0], indexes[1][1]):
                for d3idx in range(indexes[2][0], indexes[2][1]):
                    indices.append([d1idx, d2idx, d3idx])

    elif num_dims == 4:
        for d1idx in range(indexes[0][0], indexes[0][1]):
            for d2idx in range(indexes[1][0], indexes[1][1]):
                for d3idx in range(indexes[2][0], indexes[2][1]):
                    for d4idx in range(indexes[3][0], indexes[3][1]):
                        indices.append([d1idx, d2idx, d3idx, d4idx])
    else:
        raise ValueError('Not implemented for rank >2')

    # cprint(indexes, 'blue')
    # cprint(indices, 'yellow')
    # cprint(updates.shape, 'green')

    return tf.tensor_scatter_nd_update(dst_tensor,
                                       indices=indices,
                                       updates=updates)


# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from evoflow.utils import slices2array
from evoflow.engine import OP
from evoflow import backend as B


class Reverse(OP):

    O_AUTOGRAPH = True
    O_XLA = False

    def __init__(self, population_fraction, max_reverse_probability, **kwargs):
        """Reverse part of the genes.

        Args:
            population_fraction (float): How many chromosomes
            should have some of their genes reversed.

            max_reverse_probability (list(float)): What is the maximum fraction
            of the chromosome that should be affected by the reverse. Number is
            drawon from [0, max_reverse_probability]

            debug (bool, optional): print debug information and function
            returns additional data.

        Returns:
            tensor: population with some genes reversed.

        See:
            https://arxiv.org/pdf/1901.05737.pdf
        """

        if not (0 < population_fraction <= 1.0):
            raise ValueError("population_fraction must be in (0. 1]")

        for val in max_reverse_probability:
            if not (0 < val <= 1.0):
                raise ValueError(
                    "max_reverse_probability values must be between (0. 1]")

        self.population_fraction = population_fraction
        self.max_reverse_probability = max_reverse_probability
        self.has_cache = False
        self.mutation_shape = []
        self.slices = None
        super(Reverse, self).__init__(**kwargs)

    def call(self, population):

        if not self.debug:
            population = B.shuffle(population)

        # how many chromosomes to crossover
        num_reversed_chromosomes = int(population.shape[0] *
                                       self.population_fraction)

        self.print_debug('num chromosomes', num_reversed_chromosomes)

        # compute the shape needed for the mutation
        mutations_shape = [num_reversed_chromosomes]
        for idx, frac in enumerate(self.max_reverse_probability):
            max_genes = int(population.shape[idx + 1] * frac + 1)
            # ! not an error: reverse need at least 2 indices to make sense.
            if max_genes > 2:
                num_genes = B.randint(2, high=max_genes)
            else:
                num_genes = 2
            mutations_shape.append(num_genes)
            self.print_debug(idx, 'num_genes', num_genes, 'max', max_genes)

        self.print_debug("population_shape:", population.shape)
        self.print_debug("mutation_shape:", mutations_shape)

        # compute the fancy indexing dynamlically
        # ! the start point must be randomized
        slices = [slice(0, num_reversed_chromosomes)]
        for idx, crossover_size in enumerate(mutations_shape[1:]):
            # ! making indexing explicit as its a huge pitfall
            mutation_dim = idx + 1
            max_start = population.shape[mutation_dim] - crossover_size + 1
            start = B.randint(0, max_start)
            # start = random.randint(0, max_start)
            slices.append(slice(start, crossover_size + start))
        slices = tuple(slices)
        tslices = slices2array(slices)
        self.print_debug('slices', slices)

        # revesing
        reversed_population = population[slices]
        axis = B.tensor([x for x in range(1, len(reversed_population.shape))])
        reversed_population = B.reverse(reversed_population, axis)
        self.print_debug('reversed population', reversed_population)

        # assigning
        population = B.assign(population, reversed_population, tslices)

        return population


class Reverse1D(Reverse):
    def __init__(self,
                 population_fraction=0.9,
                 max_reverse_probability=0.2,
                 **kwargs):
        if not isinstance(max_reverse_probability, float):
            raise ValueError('max_reverse_probability must be a float')

        super(Reverse1D,
              self).__init__(population_fraction=population_fraction,
                             max_reverse_probability=[max_reverse_probability],
                             **kwargs)


class Reverse2D(Reverse):
    def __init__(self,
                 population_fraction=0.9,
                 max_reverse_probability=(0.2, 0.2),
                 **kwargs):

        if len(max_reverse_probability) != 2:
            raise ValueError('max_reverse_probability must be of form (x, y)')
        super(Reverse2D,
              self).__init__(population_fraction=population_fraction,
                             max_reverse_probability=max_reverse_probability,
                             **kwargs)


class Reverse3D(Reverse):
    def __init__(self,
                 population_fraction=0.9,
                 max_reverse_probability=(0.2, 0.2, 0.2),
                 **kwargs):

        if len(max_reverse_probability) != 3:
            raise ValueError(
                'max_reverse_probability must be of form (x, y, z)')

        super(Reverse3D,
              self).__init__(population_fraction=population_fraction,
                             max_reverse_probability=max_reverse_probability,
                             **kwargs)


if __name__ == '__main__':
    from copy import copy
    from termcolor import cprint
    from evoflow.utils import op_optimization_benchmark

    NUM_RUNS = 3  # 100
    # pop_shape = (100, 100, 100)
    pop_shape = (100, 100, 100)
    population = B.randint(0, 256, pop_shape)
    population_fraction = 0.5
    max_reverse_probability = (0.5, 0.5)

    OP = Reverse2D(population_fraction, max_reverse_probability)
    op_optimization_benchmark(population, OP, NUM_RUNS).report()
    quit()

    GENOME_SHAPE = (6, 4)
    population = B.randint(0, 256, GENOME_SHAPE)
    population_fraction = 0.5
    max_reverse_probability = 0.5
    cprint(population, 'green')
    original_population = copy(population)
    # ! population will be shuffle if not debug
    population = Reverse1D(population_fraction,
                           max_reverse_probability,
                           debug=True)(population)

    cprint(population, 'blue')
    # diff matrix
    diff = B.clip(abs(population - original_population), 0, 1)
    print(diff)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import evoflow.backend.tensorflow as B

# from evoflow.backend.tf_ops.assign import assign as assign2
from evoflow.backend.common import intx
from evoflow.utils import slices2array
import tensorflow as tf
import numpy as np
from perfcounters import PerfCounters

# @tf.function()
# def assign_tf(dst_tensor, updates, indexes):
#     return B.assign(dst_tensor, updates, indexes)


@tf.function()
def assign_tf(dst_tensor, values, slices):
    return B.assign(dst_tensor, values, slices)


@tf.function(experimental_compile=True)
def assign_xla(dst_tensor, values, slices):
    return B.assign(dst_tensor, values, slices)


def test_assign_fn():
    NUM_TESTS = 3
    t = np.random.randint(0, 100, (100, 100))
    dst_tensor = np.zeros(t.shape, dtype=intx())
    slices = (slice(1, 30), slice(1, 25))
    indexes = slices2array(slices)
    updates = B.cast(t[slices], dtype=intx())

    r = B.assign(dst_tensor, updates, indexes)
    r2 = assign_tf(dst_tensor, updates, indexes)
    assert B.tensor_equal(r, r2)

    assign_tf(dst_tensor, updates, indexes)
    assign_xla(dst_tensor, updates, indexes)

    cnts = PerfCounters()
    cnts.start('basic')
    for _ in range(NUM_TESTS):
        B.assign(dst_tensor, updates, indexes)
    cnts.stop('basic')

    cnts.start('tf_fn')
    for _ in range(NUM_TESTS):
        assign_tf(dst_tensor, updates, indexes)
    cnts.stop('tf_fn')

    cnts.start('tf_xla')
    for _ in range(NUM_TESTS):
        assign_xla(dst_tensor, updates, indexes)
    cnts.stop('tf_xla')

    cnts.report()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from evoflow.utils import slices2array
from evoflow.engine import OP
from evoflow import backend as B


class DualCrossover(OP):

    O_AUTOGRAPH = True
    O_XLA = False

    def __init__(self, population_fraction, max_crossover_probability,
                 **kwargs):
        """Perform Dual crossovers on a given population.

        Args:
            population_fraction (float): How many chromosomes
            should have a cross-over.

            max_crossover_probability (list(float)):  What is the maximum
            fraction of the genes that will be affected by the crossover.

            debug (bool, optional): print debug information and function
            returns additional data.

        Returns:
            tensor: population with a crossover.

        See:
            https://en.wikipedia.org/wiki/Crossover_(genetic_algorithm)
        """

        if not (0 < population_fraction <= 1.0):
            raise ValueError("population_fraction must be in ]0. 1]")

        for val in max_crossover_probability:
            if not (0 < val <= 1.0):
                raise ValueError(
                    "max_crossover_probability values must be between ]0. 1]")

        self.population_fraction = population_fraction
        self.max_crossover_probability = max_crossover_probability
        super(DualCrossover, self).__init__(**kwargs)

    def call(self, population):

        # mix genomes
        shuffled_population = B.copy(population)
        shuffled_population = B.shuffle(shuffled_population)

        # how many chromosomes to crossover
        num_crossover_chromosomes = int(population.shape[0] *
                                        self.population_fraction)
        self.print_debug('num chromosomes', num_crossover_chromosomes)

        # compute the shape needed for the mutation
        mutations_shape = [num_crossover_chromosomes]
        for idx, frac in enumerate(self.max_crossover_probability):
            max_genes = int(population.shape[idx + 1] * frac + 1)
            if max_genes > 1:
                num_genes = B.randint(1, high=max_genes)
            else:
                num_genes = 1
            mutations_shape.append(num_genes)
        mutations_shape = mutations_shape
        self.print_debug("population_shape:", population.shape)
        self.print_debug("mutation_shape:", mutations_shape)

        # compute the fancy indexing dynamlically
        # ! the start point must be randomized
        slices = [slice(0, num_crossover_chromosomes)]
        for idx, crossover_size in enumerate(mutations_shape[1:]):
            # ! making indexing explicit as its a huge pitfall
            mutation_dim = idx + 1
            max_start = population.shape[mutation_dim] - crossover_size + 1
            start = B.randint(0, max_start)
            slices.append(slice(start, crossover_size + start))
        slices = tuple(slices)
        tslices = slices2array(slices)
        self.print_debug('slices', slices)

        # crossover
        cross_section = shuffled_population[slices]
        population = B.assign(population, cross_section, tslices)

        return population


class DualCrossover1D(DualCrossover):
    def __init__(self,
                 population_fraction=0.9,
                 max_crossover_probability=0.2,
                 **kwargs):
        if not isinstance(max_crossover_probability, float):
            raise ValueError('max_crossover_probability must be a float')

        super(DualCrossover1D, self).__init__(
            population_fraction=population_fraction,
            max_crossover_probability=[max_crossover_probability],
            **kwargs)


class DualCrossover2D(DualCrossover):
    def __init__(self,
                 population_fraction=0.9,
                 max_crossover_probability=(0.2, 0.2),
                 **kwargs):

        if len(max_crossover_probability) != 2:
            raise ValueError(
                'max_crossover_probability must be of form (x, y)')
        super(DualCrossover2D, self).__init__(
            population_fraction=population_fraction,
            max_crossover_probability=max_crossover_probability,
            **kwargs)


class DualCrossover3D(DualCrossover):
    def __init__(self,
                 population_fraction=0.9,
                 max_crossover_probability=(0.2, 0.2, 0.2),
                 **kwargs):

        if len(max_crossover_probability) != 3:
            raise ValueError(
                'max_crossover_probability must be of form (x, y, z)')

        super(DualCrossover3D, self).__init__(
            population_fraction=population_fraction,
            max_crossover_probability=max_crossover_probability,
            **kwargs)


if __name__ == '__main__':
    from copy import copy
    from evoflow.utils import op_optimization_benchmark
    from termcolor import cprint
    NUM_RUNS = 3
    pop_shape = (100, 200, 100)
    population = B.randint(0, 256, pop_shape)
    population_fraction = 0.5
    max_reverse_probability = (0.5, 0.5)

    OP = DualCrossover2D(population_fraction,
                         max_reverse_probability,
                         optimization_level=0)
    op_optimization_benchmark(population, OP, NUM_RUNS).report()
    quit()
    GENOME_SHAPE = (6, 4, 4)
    population = B.randint(0, 256, GENOME_SHAPE)
    population_fraction = 0.5
    max_crossover_size_fraction = (0.5, 0.5)
    print(population.shape)
    original_population = copy(population)
    population = DualCrossover2D(population_fraction,
                                 max_crossover_size_fraction,
                                 debug=True)(population)

    # diff matrix
    diff = B.clip(abs(population - original_population), 0, 1)
    print(diff)



from evoflow.engine import EvoFlow
from evoflow.ops import Input, RandomMutations2D, UniformCrossover2D
from evoflow.selection import SelectFittest
from evoflow.fitness import Sum
from evoflow.population import randint_population
import evoflow.backend as B


def test_direct_2d():
    NUM_EVOLUTIONS = 2
    POPULATION_SIZE = 3
    GENE_SIZE = 4
    MAX_VAL = 10
    SHAPE = (POPULATION_SIZE, GENE_SIZE, GENE_SIZE)
    population = randint_population(SHAPE, MAX_VAL)
    fitness_function = Sum(max_sum_value=GENE_SIZE)
    evolution_strategy = SelectFittest()

    inputs = Input(shape=SHAPE)
    # x = RandomMutations2D(max_gene_value=1, min_gene_value=0)(inputs)
    outputs = UniformCrossover2D()(inputs)

    ef = EvoFlow(inputs, outputs, debug=True)
    ef.compile(evolution_strategy, fitness_function)
    ef.evolve(population, generations=NUM_EVOLUTIONS, verbose=0)


if __name__ == "__main__":
    test_direct_2d()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf
from tensorflow.errors import InvalidArgumentError

from .common import _infer_dtype
from evoflow.config import intx, floatx
import numpy

from evoflow.backend.tf_ops.assign import assign  # noqa: F401
from evoflow.backend.tf_ops.assign import assign2  # noqa: F401

# - tf specific functions
# ! don't import them in __init__
RAND_GENERATOR = tf.random.get_global_generator()


def get_num_gpu():
    "Return the number of GPU available to Tensorflow"
    return len(tf.config.get_visible_devices('GPU'))


# -initialization-
def tensor(a, dtype=None):
    """Converts an object to a tensor.

    Args:
        a: The source object.
        dtype: Data type specifier. It is inferred from the input by default.
    Returns:
        ndarray: An array on the current device. If ``a`` is already on
        the device, no copy is performed.

    """
    if is_tensor(a):
        return a

    if not dtype and not isinstance(a, numpy.ndarray):
        # automatic inference based of floatx and intx settings
        dtype = _infer_dtype(a)

    return tf.convert_to_tensor(a, dtype=dtype)


def copy(tensor):
    """Copy a tensor

    Args:
        tensor (ndarray): tensor to copy
    Returns
        ndarray: copied tensor
    """
    return tf.identity(tensor)


def zeros(shape, dtype=float):
    """Returns a new Tensor of given shape and dtype, filled with zeros.
    Args:
        shape (int or tuple of ints): Dimensionalities of the array.
        dtype (str, optional): dtype: Data type specifier. Defaults to 'l'.

    Returns:
        ndarray: An tensor filled with zeros.

    """
    return tf.zeros(shape, dtype=dtype)


def ones(shape, dtype=float):
    """Returns a new Tensor of given shape and dtype, filled with ones.
    Args:
        shape (int or tuple of ints): Dimensionalities of the array.
        dtype: Data type specifier.

    Returns:
        ndarray: An tensor filled with zeros.

    """
    return tf.ones(shape, dtype=dtype)


def fill(shape, fill_value):
    """Returns a new Tensor of given shape and dtype, filled with the provided
    value.

    Args:
        shape (int or tuple of ints): Dimensionalities of the array.
        fill_value (int): The value to fill with.

    Returns:
        ndarray: An tensor filled with zeros.

    """
    return tf.fill(shape, fill_value)


def normal(shape, mean=0, stddev=1.0):
    """Draw random samples from a normal (Gaussian) distribution.

    Args:
        shape (int or tuple of ints): Dimensionalities of the array.
        mean (float): Mean value. Default to 0.0.
        stddev (float): Standard deviations. Default to 1.0.

    Returns:
        ndarray: Drawn samples from the parameterized normal distribution.

    """
    return RAND_GENERATOR.normal(shape, mean=mean, stddev=stddev)


def range(start, stop=None, step=1, dtype=intx()):
    """Creates a sequence of numbers.
    Creates a sequence of numbers that begins at `start` and extends by
    increments of `delta` up to but not including `stop`.

    Args:
        start (int): Initial value. Optional. Defaults to 0
        stop (int, optional): End value.
        delta (int, optional): Spacing between values.  Defaults to 1.
        dtype (str, optional): Tensor tyoe. Defaults to intx().

    Returns:
        Tensor: Tensor that contains the requested range.
    """
    return tf.range(start, limit=stop, delta=step, dtype=dtype)


# - Reduce -
def prod(tensor, axis=None, keepdims=False):
    """Returns the product of an array along a given axis.

    Args:
        tensor (ndarray): Array to take the maximum.
        axis (int): Along which axis to take the maximum. The flattened array
            is used by default. Defaults to None.
        keepdims (bool): If ``True``, the axis is kept as an axis of
        size one. Default to False.

    Returns:
        ndarray: The maximum of ``tensor``, along the axis if specified.
    """
    return tf.math.reduce_prod(tensor, axis=axis, keepdims=keepdims)


def max(tensor, axis=None, keepdims=False):
    """Returns the maximum of an array or the maximum along a given axis.

    Note::
       When at least one element is NaN, the corresponding min value will be
       NaN.

    Args:
        tensor (ndarray): Array to take the maximum.
        axis (int): Along which axis to take the maximum. The flattened array
            is used by default. Defaults to None.
        keepdims (bool): If ``True``, the axis is kept as an axis of
        size one. Default to False.

    Returns:
        ndarray: The maximum of ``tensor``, along the axis if specified.
    """

    return tf.math.reduce_max(tensor, axis=axis, keepdims=keepdims)


def min(tensor, axis=None, keepdims=False):
    """Returns the minimum of an array or the maximum along an axis.

    Note::
       When at least one element is NaN, the corresponding min value will be
       NaN.

    Args:
        tensor (ndarray): Array to take the maximum.
        axis (int): Along which axis to take the maximum. The flattened array
                    is used by default. Defaults to None.
        keepdims (bool): If ``True``, the axis is kept as an axis of
        size one. Default to False.

    Returns:
        ndarray: The maximum of ``tensor``, along the axis if specified.
    """
    return tf.math.reduce_min(tensor, axis=axis, keepdims=keepdims)


def sum(tensor, axis=None, keepdims=False):
    """Returns the sum of an array along given axes.

    Args:
        tensor (ndarray): Array to sum reduce.
        axis (int or sequence of ints): Axes along which the sum is taken.
        keepdims (bool): If ``True``, the specified axes are remained as axes
        of length one.


    Returns:
        ndarray: The sum of ``tensor``, along the axis if specified.
    """
    return tf.math.reduce_sum(tensor, axis=axis, keepdims=keepdims)


def mean(tensor, axis=None, keepdims=False):
    """Returns the sum of an array along given axes.

    Args:
        tensor (ndarray): Array to mean reduce.
        axis (int or sequence of ints): Axes along which the sum is taken.
        keepdims (bool): If ``True``, the specified axes are remained as axes
        of length one.

    Returns:
        ndarray: The mean of ``tensor``, along the axis if specified.
    """
    return tf.math.reduce_mean(tensor, axis=axis, keepdims=keepdims)


def sqrt(tensor):
    """Computes element-wise square root of the input tensor.

    Args:
        tensor (ndarray): tensor

    Returns:
        tensor: square root of the input tensor.
    """
    return tf.math.sqrt(tensor)


# - Manipulation -
def reverse(tensor, axis):
    """Reverses specific dimensions of a tensor.

    Args:
        tensor (tensor): tensor to reverse
        axis (tensor): axis or tuple of axis
    """
    return tf.reverse(tensor, axis)


def roll(tensor, shift, axis):
    """Rolls the elements of a tensor along an axis.

    Args:
        tensor (tensor): tensor to roll
        shift (tensor): offset to roll
        axis (tensor): axis to shift by

    Returns:
        [type]: [description]
    """
    return tf.roll(tensor, shift, axis)


def tile(tensor, reps):
    """Construct a tensor by repeating tensor the number of times given by reps.
    Args:
        tensor (cupy.ndarray): Tensor to transform.
        reps (int or tuple): The number of repeats.
    Returns:
        ndarray: Transformed tensor with repeats.
    """
    return tf.tile(tensor, reps)


def concatenate(tup, axis=0):
    """Joins tensors along a given axis.
    Args:
        tup (sequence of arrays): Tensors to be joined. Tensors must have the
        same cardinality except for the specified axis.

        axis (int or None): Axis to joint along. If None, tensors are
        flattened before use. Default is 0.
    Returns:
        ndarray: Joined array.
    """
    return tf.concat(tup, axis=axis)


# - Utils -
def transpose(a):
    "Transpose the tensor"
    return tf.transpose(a)


def cast(tensor, dtype):
    """Cast

    Args:
        tensor (Tensor): tensor to cast.
        dtype (str): type to cast. Usually floatx() or intx()

    Returns:
        ndarray: Tensor casted in the requested format.
    """
    return tf.cast(tensor, dtype)


def dtype(tensor):
    """"Returns the dtype of a tensor as a string.

    Args:
        tensor (tensor): Tensor

    Returns:
        str: type of the tensor as string. e.g int32.
    """
    return tensor.dtype.name


def flatten(tensor):
    """Returns a copy of the tensor flatten into one dimension.

    Args:
        tensor (ndarray): tensor to flatten
    Returns:
        ndarray: flattened tensor
    """
    # note: unsure if that is the fastest way. maybe compute flat_shape in
    # pure python

    # flat_shape = prod(tensor.shape)
    return tf.reshape(tensor, [-1])


def as_numpy_array(t):
    """Convert tensor to a numpy array.

    Useful when interfacing with other librairies to have a unified input
    to them.

    Args:
        t (ndarray): tensor to convert
    Returns:
        numpy.ndarray: Tensor as numpy array
    """

    if not is_tensor(t):
        dtype = _infer_dtype(t)
        if isinstance(t, list):
            t = tensor(t)
        else:
            t = tensor([t])
    else:
        dtype = t.dtype.name

    return t.numpy().astype(dtype)


def reshape(tensor, shape):
    "reshape tensor"
    return tf.reshape(tensor, shape)


def is_tensor(a):
    "check if the given object is a tensor"
    if isinstance(a, tf.Tensor):
        return True
    elif isinstance(a, tf.python.framework.tensor_shape.TensorShape):
        return True
    else:
        return False


def tensor_equal(tensor1, tensor2):
    """True if two tensors are exactly equals

    Args:
        tensor1 (ndarray): tensor to compare.
        tensor2 (ndarray): tensor to compare.

    Returns:
        Bool: True if exactly equal, False otherwise.

    """
    return tf.reduce_all(tf.equal(tensor1, tensor2))


def assert_near(a, b, absolute_tolerance=0, relative_tolerance=0):
    """
    Returns True if two arrays are element-wise equal within a tolerance.

    Args:
        a (ndarray): Tensor with a last dimension of at least k size.
        b (ndarray): Tensor with a last dimension of at least k size.
        absolute_tolerance (float): Default to 0
        relative_tolerance (float): Default to 0

    Returns:
        bool: True if the two arrays are equal within the given tolerance; False otherwise.

    Note:
        This function return True if the following equation is satisfied:
        `absolute(a - b) <= (absolute_tolerance + relative_tolerance * absolute(b))`  # noqa

    """

    try:
        tf.debugging.assert_near(a,
                                 b,
                                 atol=absolute_tolerance,
                                 rtol=relative_tolerance)
    except InvalidArgumentError:
        return False
    return True


# - Math -
def _is_scalar(a):
    if isinstance(a, int):
        return True
    elif isinstance(a, float):
        return True
    elif is_tensor(a):
        if a.shape == ():
            return True
    else:
        False


def dot(t1, t2):
    """Return the dot product of two arrays

    Args:
        t1 (ndarray): Left tensor
        t2 (ndarray): Right tensor

    Return:
        ndarray: tensor containing the dot product
    """

    # # scalar
    if (_is_scalar(t1) or _is_scalar(t2)):
        return t1 * t2
    #     return tf.reduce_sum(tf.multiply(t1, t2))
    elif len(t1.shape) == 1 or len(t2.shape) == 1:
        return tf.reduce_sum(tf.multiply(t1, t2), axis=-1)
    else:
        return tf.matmul(t1, t2)


def add(tensor1, tensor2):
    """Add two tensors

    Args:
        tensor1 (ndarray): Left tensor.
        tensor2 (ndarray): right tensor.
    """
    return tf.add(tensor1, tensor2)


def subtract(tensor1, tensor2):
    """Substract two tensors

    Args:
        tensor1 (ndarray): Left tensor.
        tensor2 (ndarray): right tensor.
    """
    return tf.subtract(tensor1, tensor2)


def multiply(tensor1, tensor2):
    """multiply two tensors

    Args:
        tensor1 (ndarray): Left tensor.
        tensor2 (ndarray): right tensor.
    """
    return tf.multiply(tensor1, tensor2)


def divide(numerator, denominator):
    """divide a tensor by another

    Args:
        tensor1 (ndarray): numerator tensor.
        tensor2 (ndarray): denominator tensor.
    """
    return tf.cast(tf.divide(numerator, denominator), floatx())


def mod(numerator, denominator):
    """Compute the reminder of the divisin of a tensor by another

    Args:
        tensor1 (ndarray): numerator tensor.
        tensor2 (ndarray): denominator tensor.
    """
    return tf.math.mod(numerator, denominator)


def clip(tensor, min_val=0, max_val=None, out=None):
    """Clips the values of a tensor to a given interval. For example,
    if an interval of [0, 1] is specified, values smaller than 0 become 0,
    and values larger than 1 become 1.

    Efficient version of ``max(min(a, max_val), min_val)``

    Args:
        tensor (ndarray): The input Tensor.

        min_val (scalar, ndarray or None): The left side of the interval. When
        None ignored.  Defaults to None.

        max_val (scalar, ndarray or None): The right side of the interval. When
        None ignored.  Defaults to None.

    Returns:
        ndarray: Clipped tensor.
    """
    return tf.clip_by_value(tensor,
                            clip_value_min=min_val,
                            clip_value_max=max_val)


def abs(tensor):
    "Calculate the absolute value element-wise."
    return tf.math.abs(tensor)


def broadcasted_norm(tensor):
    "Norm broadcasted accross dimensions"
    norm = cast(tf.abs(tensor), intx())
    norm = norm**2
    norm = tf.reduce_sum(norm, axis=-1)
    norm = sqrt(cast(norm, floatx()))
    return norm

    # population_norm = B.sum(B.abs(flat_pop)**2, axis=-1)**0.5


def norm(tensor, ord='euclidean', axis=None, keepdims=False):
    """Return one of eight different matrix norms, or one of an infinite
    number of vector norms (described below), depending on the value of
    the `ord` parameter.

    Args:
        tensor (ndarray): Array to take the norm from. If ``axis`` is None,
        the ``tensor`` must be 1D or 2D.

        ord (non-zero int, inf, -inf, 'fro'): Norm type. Euclidian by default.
        See `tf.norm` for explanation:
        https://www.tensorflow.org/api_docs/python/tf/norm

        axis (int, 2-tuple of ints, None): `axis` along which the norm is
        computed.

        keepdims (bool): If this is set ``True``, the axes which are normed
        over are left in the resulting tensor with a size of one.

    Returns:
        ndarray: Norm of the tensor.
    """
    return tf.norm(tensor, ord=ord, axis=axis, keepdims=keepdims)


# - Randomness -


def randint(low, high=0, shape=None, dtype=intx()):
    """Returns a scalar or an array of integer values over [low, high)

    Args:
        low (int): If high is None, it is the upper bound of the
        interval and the lower bound is set to 0. if high is set, it is the
        lower bound of the interval.

        high (int, optional):Upper bound of the interval. Defaults to None.

        shape (None or int or tuple of ints, optional): The shape of returned
        value. Defaults to None.

        dtype (str, optional): dtype: Data type specifier.
        Defaults to 'float32'.

    Returns:
        int or ndarray of ints: If size is None, it is single integer
        sampled. If size is integer, it is the 1D-array of length size
        element. Otherwise, it is the array whose shape specified by size.
    """

    # just one number
    if not shape:
        if high == 0:
            high = low
            low = 0
        return RAND_GENERATOR.uniform(shape=(1, ),
                                      minval=low,
                                      maxval=high,
                                      dtype=intx())[0]

    if isinstance(shape, int):
        shape = (shape, )
    return RAND_GENERATOR.uniform(shape=shape,
                                  minval=low,
                                  maxval=high,
                                  dtype=dtype)


def shuffle(t, axis=0):
    """Shuffle tensor along the given axis. Other axis remain in
    place.

    Args:
        tensor (ndarray): tensor to shuffle.
        axis (int, optional): axis to shuffle on. Default to 0.

    Returns:
        None: in place shuffling
    """
    if not axis:
        # ! tensorflow don't do in place shuffling
        return tf.random.shuffle(t)
    else:
        # FIXME: its a hack as we use numpy which is liklely to cause slowness
        t = as_numpy_array(t)
        rng = numpy.random.default_rng()
        rng.shuffle(t, axis=axis)
        return tensor(t)


def full_shuffle(t):
    """Shuffle in place a tensor along all of its axis

    Args:
        t (ndarray): tensor to shuffle.


    Returns:
        tensor: shuffled tensor


    We need to use transpose

    Shuffle()

    2D
    [1, 0] -> [1, 0] : Dim 1
    [1, 0] -> [0, 1] : restore order

    3D
    Transpose -> new order
    [1, 0, 2] -> [1, 0, 2]  : Dim 1
    [2, 1, 0] -> [2, 0, 1]  : Dim 2
    [1, 2, 0] -> [0, 1, 2]  : restore order

    4D
    Transpose -> new order
    [1, 0, 2, 3] -> [1, 0, 2, 3]  : Dim 1
    [2, 1, 0, 3] -> [2, 0, 1, 3]  : Dim 2
    [3, 1, 2, 0] -> [3, 0, 1, 2]  : Dim 3
    [1, 2, 3, 0] -> [0, 1, 2, 3]  : restore order

    5D+
    nope - might consider patching tf.shuffle at that point :)

    """
    # ! dont use the variable name tensor as it confusion with the tensor()

    dims = len(t.shape)

    t = tf.random.shuffle(t)  # always shuffle dim0
    if dims == 2:

        # dim 1
        t = tf.transpose(t, [1, 0])
        t = tf.random.shuffle(t)

        # restore
        t = tf.transpose(t, [1, 0])

    elif dims == 3:

        # dim 1
        t = tf.transpose(t, [1, 0, 2])
        t = tf.random.shuffle(t)

        # dim 2
        t = tf.transpose(t, [2, 1, 0])
        t = tf.random.shuffle(t)

        # restore
        t = tf.transpose(t, [1, 2, 0])

    elif dims == 4:

        # dim 1
        t = tf.transpose(t, [1, 0, 2, 3])
        t = tf.random.shuffle(t)

        # dim 2
        t = tf.transpose(t, [2, 1, 0, 3])
        t = tf.random.shuffle(t)

        # dim 3
        t = tf.transpose(t, [3, 1, 2, 0])
        t = tf.random.shuffle(t)

        # restore
        t = tf.transpose(t, [1, 2, 3, 0])

    elif dims > 4:
        print('tensor shape', t.shape)
        raise ValueError('Tensor Rank > 4 -- not implemented')

    return t
    # FIXME: its a hack as we use numpy which is liklely to cause slowness
    # t = as_numpy_array(t)
    # rng = numpy.random.default_rng()
    # for idx in range(len(t.shape)):
    #     rng.shuffle(t, axis=idx)
    # return tensor(t)


# - Indexing -
def take(tensor, indices, axis=None):
    """Takes elements of a Tensor at specified indices along a specified axis

    Args:
        tensor (ndarray): Tensor to extract elements from.

        indices (int or array-like): Indices of elements that this function
        takes.

        axis (int, optional): The axis along which to select indices from.
        The flattened input is used by default. Defaults to None.

    Returns:
        ndarray: Tensor containing the values from the specified indices.
    """
    return tf.gather(tensor, indices, axis=axis)


def top_k_indices(tensor, k, axis=-1):
    """
    Finds the indices of the k largest entries alongside an axis.

    Args:
        tensor (ndarray): Tensor with a last dimension of at least k size.

        k (i): number of elements to return

    """
    return tf.math.top_k(tensor, k)[1]


def bottom_k_indices(tensor, k, axis=-1):
    """
    Finds the indices of the k smallest entries alongside an axis.

    Args:
        tensor (ndarray): Tensor with a last dimension of at least k size.

        k (i): number of elements to return.

        axis (int or None) - Axis along which to sort. Default is -1,
        which is the last axis. If None is supplied,
        the array will be flattened before sorting.

    """
    # inverted top k and reinverted before returning
    return tf.math.top_k(tf.multiply(tensor, -1), k)[1]


def unique_with_counts(tensor):
    """Finds unique elements and return them along side their position and
    counts.

    Args:
        tensor (Tensor): 1D tensor to analyze.

    Returns:
        values (Tensor): unique values founded
        indexes (Tensor): index of the value sorted
        counts (Tensor): Tensor containing the count for each value.
    """
    return tf.unique_with_counts(tensor)


# - Statistical -


def bincount(tensor, weights=None, minlength=None):
    """Count number of occurrences of each value in array of non-negative ints.

    Args:
        tensor (ndarray): Input tensor.

        weights (cupy.ndarray): Weights tensor which has the same shape as
        tensor``. Default to None.

        minlength (int): A minimum number of bins for the output array.

    """
    return tf.math.bincount(tensor, weights=weights, minlength=minlength)



import tensorflow as tf

RAND_GENERATOR = tf.random.get_global_generator()


@tf.function()
def randint2(low, high=None, shape=None, dtype='int32'):
    """Returns a scalar or an array of integer values over [low, high)

    Args:
        low (int): If high is None, it is the upper bound of the
        interval and the lower bound is set to 0. if high is set, it is the
        lower bound of the interval.

        high (int, optional):Upper bound of the interval. Defaults to None.

        shape (None or int or tuple of ints, optional): The shape of returned
        value. Defaults to None.

        dtype (str, optional): dtype: Data type specifier.
        Defaults to 'float32'.

    Returns:
        int or ndarray of ints: If size is None, it is single integer
        sampled. If size is integer, it is the 1D-array of length size
        element. Otherwise, it is the array whose shape specified by size.
    """

    # just one number
    if not shape:
        return RAND_GENERATOR.uniform(shape=(1, ),
                                      minval=low,
                                      maxval=high,
                                      dtype='int32')[0]

    if isinstance(shape, int):
        shape = (shape, )
    return RAND_GENERATOR.uniform(shape=shape,
                                  minval=low,
                                  maxval=high,
                                  dtype=dtype)
